
(*open Syntax*)

(* type ('a,'b) arg_list = *)
(*   | End : ('a,'a) arg_list *)
(*   | Arg : 'a * ('b,'c) arg_list -> ('a -> 'b, 'c) arg_list *)



  type types = 
    | Tcs
    | T_bool
    | T_int
	
  type funcs =
    | Plus
    | Minus
    | Times
    | Theta_power
    | Helper_power
    | Theta_fact
    | Helper_fact
    | Theta_fib
    | Helper_fib
    | Theta_mul
    | Helper_mul
    | Theta_expt
    | Helper_expt
    | Theta_less
    | Helper_less
    | Theta_sum_square
    | Helper_sum_square
    | Theta_sum
    | Helper_sum
    | Theta_example
    | Helper_example
	
  type preds =
    | Equals 
	
  type term = 
    | T_atom of int
    | T_var of string
    | T_func of funcs * (term list)
	
(* NEED TO GENERALISE INT HERE TO TYPE *)

  type form = 
    | F_comparison of preds * (term list)
	
	
  (* import syntax structure from logic *)
	
 

let get_tc typ =
  match typ with
    | T_int ->
      1::2::3::4::[]
    | _ ->
      []

let pretty_func f =
  match f with
    | Times -> "*"
    | Plus -> "+"
    | Minus -> "-"
    | Theta_power -> "theta_power"
    | Helper_power -> "helper_power"
    | Theta_fact -> "theta_fact"
    | Helper_fact -> "helper_fact"
    | Theta_fib -> "theta_fib"
    | Helper_fib -> "helper_fib"
    | Theta_mul -> "theta_mul"
    | Helper_mul -> "helper_mul"
    | Theta_expt -> "theta_expt"
    | Helper_expt -> "helper_expt"
    | Theta_less -> "theta_less"
    | Helper_less -> "helper_less"
    | Theta_sum_square -> "theta_sum_square"
    | Helper_sum_square -> "helper_sum_square"
    | Theta_sum -> "theta_sum"
    | Helper_sum -> "helper_sum"
    | Theta_example -> "theta_example"
    | Helper_example -> "helper_example"

let get_func_types_t f =
  match f with
    | Times -> T_int::T_int::[]
    | Plus -> T_int::T_int::[]
    | Minus -> T_int::T_int::[]
    | Theta_power -> T_int::[]
    | Helper_power -> T_int::T_int::[]
    | Theta_fact -> T_int::[]
    | Helper_fact -> T_int::T_int::[]
    | Theta_fib -> T_int::[]
    | Helper_fib -> T_int::T_int::T_int::[]
    | Theta_mul -> T_int::T_int::[]
    | Helper_mul -> T_int::T_int::T_int::[]
    | Theta_expt -> T_int::T_int::[]
    | Helper_expt -> T_int::T_int::T_int::[]
    | Theta_less -> T_int::T_int::[]
    | Helper_less -> T_int::T_int::[]
    | Theta_sum_square -> T_int::[]
    | Helper_sum_square -> T_int::T_int::[]
    | Theta_sum -> T_int::[]
    | Helper_sum -> T_int::T_int::[]
    | Theta_example -> T_int::T_int::[]
    | Helper_example -> T_int::T_int::T_int::[]

(* do duplicate thing here ??? *)
let common_funcs =
    Times::Plus::Minus::[]

let rec term_pretty tt =
  match tt with
    | T_var x ->
      x
    | T_atom x ->
      string_of_int x
    | T_func(func,args) ->
      (pretty_func func)^
	(let s = (List.fold_right
		  (fun x y ->
		    (term_pretty x)^","^y) args "")
	 in ("("^(String.sub s 0 ((String.length s)-1))^")"))
	
let form_pretty t = 
  match t with
    | F_comparison(Equals,(arg1::arg2::[])) ->
      (term_pretty arg1)^" = "^(term_pretty arg2)



let rec get_test_cases typs tcs =
  let rec memin (T_var v) l = 
    match l with 
      | [] ->
	false
      | (T_var v2,_)::t ->
	if (v2 = v) then
	  true
	else
	  memin (T_var v) t
  in
  match typs with
    | [] ->
      tcs
    | (h::t) ->
      match h with
	| (T_func(_,_),_) ->
	  get_test_cases t tcs
	| (v,typ) ->
	  if memin v tcs
	  then
	    get_test_cases t tcs
	  else
	    get_test_cases t ((v,get_tc typ)::tcs)


(*--------*)


let rec theta_fact a =
      match a with
	| 0 -> 1
	| _ -> a * theta_fact (a-1)

let rec helper_fact a b = 
  match a with
    | 0 -> b
    | _ ->
      helper_fact (a-1) (a*b)

let rec times x y = 
  x*y


let rec theta_power a =
 match a with
   | 0 -> 1
   | _ -> 2*(theta_power (a-1))

let rec helper_power a b =
  match a with
    | 0 -> b
    | _ ->
      helper_power (a-1) (2*b)

let rec theta_fib a =
  if (a <= 0)
  then 
    0
  else
    (let x = (a-1) in
     (if (x<=0) 
      then 
	 1
      else
	  (theta_fib (a-1)) + (theta_fib (a-2))))
 
(* lem is tf n = hf n 0 1 *)
(* lem is hf n m a = (tf n) * m + (tf n) * a *)
let rec helper_fib n m a =
  if (n < 0) then 
    m
  else
    (match n with
      | 0 -> m
      | 1 -> a
      | _ -> helper_fib (n-1) a (m+a))
	
let rec theta_mul n m =
  match n with
    | 0 -> 0
    | _ ->
      let x = (n-1) in
      m + (theta_mul x m)

let rec helper_mul n m a =
    match n with
      | 0 -> a
      | _ ->
	helper_mul (n-1) m (a+m)

let rec theta_expt n m =
  if (m <= 0) then
    1
  else
    let x = (m-1) in
    n * (theta_expt n x)
      
let rec helper_expt n m a =
  if (m <=0) then
    a
  else
    helper_expt n (m-1) (n*a)
      
let rec theta_less n m =
  if (n<m) then 1 else 0

let rec helper_less n m =
  match (n,m) with
    | (_,0) -> 0
    | (0,_) -> 1
    | _,_ -> helper_less (n-1) (m-1)

let rec theta_sum_square n =
  match n with
    | 0 -> 0
    | _ -> (n*n) + (theta_sum_square (n-1))

let rec helper_sum_square n a =
  match n with
    | 0 -> a
    | _ -> helper_sum_square (n-1) ((n*n)+a)

let rec theta_sum n =
  (n*(n+1))/2

let rec helper_sum n a =
  match n with
    | 0 -> a
    | _ ->
      helper_sum (n-1) (n+a)

let rec theta_example x y =
  match y with
    | 0 -> 2*x
    | _ -> y + (3*x)

let rec helper_example x y a =
  match x with
    | 0 -> a
    | _ ->
      match y with
	| 0 -> helper_example (x-1) y (a+1)
	| _ -> helper_example (x-1) y (a+2)
(* don't think this is possible *)
(* lemma is 
helper_example x y a = a + (if (y==0) then x else (2*x)) *)

(* END OF USE *)





(* let rec apply f l =  *)
(*   match l with *)
(*     | [] -> failwith "GADT failure for 0 arity function" *)
(*     | [x] ->  *)
(*       match f  with *)
(* 	| Func f2 -> f2 x *)
(*     | h::t -> *)
(*       match f with *)
(* 	| Func f2 -> *)
(* 	  f2 h (apply f t) *)
	
(* let rec c_map : type a b. a -> (a,b) arg_list -> (a,b) arg_list = fun f ->  *)
(*   function  *)
(*     | End -> End *)
(*     | Arg(a,b) -> *)
(*       (f a)@::(c_map f b) *)

let rec eval exp =
  let rec eval_term exp_t =
    match exp_t with 
      | T_atom n ->
	n
(* do this more generically  - work out sequential application....*)
      | T_func(Plus,arg1::arg2::[]) ->
      	(eval_term arg1) + (eval_term arg2)
      | T_func(Minus,arg1::arg2::[]) ->
	(let ans = 
      	  (eval_term arg1) - (eval_term arg2) in
	 if (ans < 0) then 
	   0 
	 else 
	   ans
	)
      | T_func(Theta_power,arg1::[]) ->
      	theta_power (eval_term arg1)
      | T_func(Helper_power,arg1::arg2::[]) ->
      	helper_power (eval_term arg1) (eval_term arg2)
      | T_func(Times,arg1::arg2::[]) ->
      	(eval_term arg1) * (eval_term arg2)
      | T_func(Theta_fact,arg1::[]) ->
      	theta_fact (eval_term arg1)
      | T_func(Helper_fact,arg1::arg2::[]) ->
      	helper_fact (eval_term arg1) (eval_term arg2)
      | T_func(Theta_fib,arg1::[]) ->
      	theta_fib (eval_term arg1)
      | T_func(Helper_fib,arg1::arg2::arg3::[]) ->
      	helper_fib (eval_term arg1) (eval_term arg2) (eval_term arg3)
      | T_func(Theta_mul,arg1::arg2::[]) ->
      	theta_mul (eval_term arg1) (eval_term arg2)
      | T_func(Helper_mul,arg1::arg2::arg3::[]) ->
      	helper_mul (eval_term arg1) (eval_term arg2) (eval_term arg3)
      | T_func(Theta_expt,arg1::arg2::[]) ->
      	theta_expt (eval_term arg1) (eval_term arg2)
      | T_func(Helper_expt,arg1::arg2::arg3::[]) ->
      	helper_expt (eval_term arg1) (eval_term arg2) (eval_term arg3)
      | T_func(Theta_less,arg1::arg2::[]) ->
      	theta_less (eval_term arg1) (eval_term arg2)
      | T_func(Helper_less,arg1::arg2::[]) ->
      	helper_less (eval_term arg1) (eval_term arg2) 
      | T_func(Theta_sum_square,arg1::[]) ->
      	theta_sum_square (eval_term arg1)
      | T_func(Helper_sum_square,arg1::arg2::[]) ->
      	helper_sum_square (eval_term arg1) (eval_term arg2) 
      | T_func(Theta_sum, arg1::[]) ->
	theta_sum (eval_term arg1)
      | T_func(Helper_sum, arg1::arg2::[]) ->
	helper_sum (eval_term arg1) (eval_term arg2)
      | T_func(Theta_example, arg1::arg2::[]) ->
	theta_example (eval_term arg1) (eval_term arg2)
      | T_func(Helper_example, arg1::arg2::arg3::[]) ->
	helper_example (eval_term arg1) (eval_term arg2) (eval_term arg3)
  in
  match exp with
    | F_comparison(Equals,arg1::arg2::[]) ->
      let aa = 
	(eval_term arg1) in
      let bb = (eval_term arg2) in (aa==bb)
				
