(* decompose tree and collect symbols *)
open Eval_nats
open Tester

(* also collect symbols from one level down *)
exception T_enough of int

let max_gen = 2

let rec gen_vars v num = 
  let rec gen_list l = 
    match l with
      | 0 -> []
      | n ->
	n::(gen_list (n-1)) 
  in
  let rec find_top vars ans = 
    match vars with
      | [] ->
	ans
      | (T_var x)::t ->
	if x > ans then
	  find_top t x
	else
	  find_top t ans
  in
  match v with
    | [] ->
      [(T_var "a")]
    | (T_var x)::_ ->
      if (num < 0) then 
	[]
      else
	let top = find_top v x in
	List.fold_left (fun x y -> (T_var (top^(string_of_int y)))::x) [] (gen_list num)
    | x::_ ->
      [x]

let rec remdup al =
  match al with
    | [] ->
	[]
    | h::[] ->
      [h]
    | h::t ->
      if (List.mem h t)
      then 
	remdup t
      else
	h::(remdup t)

let rec get_vars_term expr = 
  match expr with 
    | T_atom n ->
      []
    | T_var x ->
      [T_var x]
    | T_func(f,args) ->
      (List.fold_left
	 (fun x y -> List.append (get_vars_term y) x) [] args)

let rec get_vars exp =
  match exp with
    | F_comparison(Equals,args) ->
      List.fold_left (fun x y -> List.append (get_vars_term y) x) [] args
	
let rec collect exp =
  let rec collect_term expr = 
    match expr with 
      | T_atom n ->
	[]
      | T_var x ->
	[]
      | T_func(f,args) ->
	f::(List.fold_left
	      (fun x y -> List.append (collect_term y) x) [] args)
  in
  match exp with
    | F_comparison(Equals,args) ->
      List.fold_left (fun x y -> List.append (collect_term y) x) [] args
(*
  %%% make decisions about symbol swapping
%%% add heuristics here?
*)

(* bdf termination of collection here? *)
(* need to form a queue *)

(* algo heuristics ---  compare arities? any more *)
(* compute all possibilities and arrange? *)
(* need to think about typing? *)

  

(* what to do here if extra args are required? *)
let group_plausible s1 s2 =
  ((List.mem s1 common_funcs) && (List.mem s2 common_funcs)) or
    ((not (List.mem s1 common_funcs)) && (not (List.mem s2 common_funcs)))
  (*  (List.length (get_func_types_t s1)) *)
  (* >=  *)
  (*   (List.length (get_func_types_t s2)) *)



(* need to work out here how to raise error when not good enough *)
(* heuristics go here *) 
(* what if lists aren't same length *)
(* source symb then potential symbs *)
let rec get_symb s ss = 
  match ss with
    | h::t ->
      (if group_plausible h s then
	  (h,t)
      else
	  ((fst (get_symb s t)),(h::(snd (get_symb s t)))))
    | [] ->
      raise (T_enough 0)
(* pick one from available symbols with heuristic choice *)    
(* return pair of (choice, rest of tail symbs) *)
(* GERHARD SUGGESTION - HASH *)
let rec find_one_equiv s ss = 
  (* let rec filter_symbs a aes ts =  *)
  (*   match aes with  *)
  (*     | [] -> [] *)
  (*     | (h,h2)::t -> *)
  (* 	(if (h=a) then *)
  (* 	  filter_symbs h t ts *)
  (* 	else *)
  (* 	  (if (List.mem h2 ts) *)
  (* 	   then  *)
  (* 	      h2::(filter_symbs h t ts) *)
  (* 	   else  *)
  (* 	      filter_symbs h t ts)) *)
  (* in *)
  let rec filter_symbs a aes =
    match aes with
      | [] -> []
      | h::t ->
	if (h=a) then
	  t
	else
	  h::(filter_symbs a t)
  in	  
 (* let symbs = filter_symbs s ss in*)
(*  get_symb s symbs*)
  get_symb s ss


let rec insert a b =
  match b with 
    | [] ->
      [[a]]
    | h::t ->
      (a::(h::t))::(List.map (fun x -> h::x) (insert a t))
	


let rec pa l =
  match l with
    | [] -> [[]]
  | h::t ->
    (match h with
      | [] -> []
      | h1::t1 ->
	let nl = (List.fold_left (fun x y -> List.append (insert h1 y) x) [] (pa t)) in
	let nr = (pa (t1::t)) in
	let aa = (List.append nl nr) in
    aa)
	
let pa3 ll = 
  (* let l =  (List.map (fun x -> remdup x) ll) in  *)
  (*  print_string "\n\n"; *)
  (* print_string (List.fold_right (fun x y -> ( *)
  (*   ("; ["^( *)
  (*     List.fold_right (fun z t -> *)
  (*     (((term_pretty z)^","^(t)))) x "") *)
  (* 		     ^"]") *)
  (*   ^(y))) l ""); *)
  print_string "\n\n";
  List.iter (fun x -> print_string ("len:"^(string_of_int (List.length x))^"  ")) ll;
  pa ll

(* do this with exception *)
let rec mem_var exp l = 
  match exp with
    | (T_var x) ->
      (match l with
	| (T_var y)::t ->
	  if (x=y)
	  then 
	    true
	  else
	    mem_var exp t
	| T_func(_,args)::t ->
	  mem_var exp args
	  or 
	    mem_var exp t
	| _ ->
	  false)
    | _ ->
      false

	
let rec pa2 l =
  match l with
    | [] -> [[]]
  | h::t ->
    (match h with
      | [] -> []
      | h1::t1 ->
	let nl = (List.fold_left (fun x y -> List.append (if (mem_var h1 y) then [] else (insert h1 y)) x) [] (pa2 t)) in
	let nr = (pa2 (t1::t)) in
	List.append nl nr)
			      
let rec pa_simp l =
  match l with
    | [] -> [[]]
  | h::t ->
    List.fold_left (fun x y -> (List.append (List.map (fun z -> z::y) h) x)) [] (pa_simp t)


let rec find_equiv res insymbs target_symbs source_symbs =
    (* List.iter (fun x -> print_string ("target: "^  (pretty_func x)^"\n\n")) target_symbs; *)
    (* List.iter (fun x -> print_string ("source: "^  (pretty_func x)^"\n\n")) source_symbs; *)
  match source_symbs with
    | [] ->
      res
    | h::t ->
      (* error clause wrap here - no good new match of symbols found *)
      let (s,rest) = find_one_equiv h target_symbs  in
      (* print_string ((pretty_func h)^"<>"^(pretty_func s)); *)
      (* print_string "\n\n"; *)
      (* List.iter (fun x -> print_string ("rest: "^  (pretty_func x)^"\n")) rest; *)
      (*  List.iter (fun x -> print_string ("t: "^(pretty_func x)^"\n")) t; *)
      find_equiv ((h,s)::res) insymbs rest t
	(* error catchall give back empty clause? *)
	(*raise T_enough n*)

(* This is the outermost function *)
let rec permute l = 
  match l with
    | [] -> [l]
    | h::t -> 
      List.flatten (List.map (fun x -> insert h x) (permute t))

(* with doubles *)
let rec permute_multi_all xs n l t =
  if (n<=0) then l::t
  else
  List.flatten (List.map (fun c -> 
    permute_multi_all xs (n-1) (c::l) t) xs)

(* without doubles*)
let rec permute_multi_all xs n l t =
  if (n<=0) then l::t
  else
     List.fold_left (fun d c -> (List.append (if (List.mem c l) then [] else permute_multi_all xs (n-1) (c::l) t) d)) [] xs;;

(* combinations *)
let rec permute_multi l n =
  if (n > List.length l)  
  then 
  []
  else
  match n with
  | 0 -> [[]]
  | n ->
  let p = (n-1) in
  match l with
    | [] -> 
      []
    | h::t ->
      (List.append
	 (List.map (fun x -> h::x) (permute_multi t p))
	 (permute_multi t n));;


(* check here what to do to get equivalences for a set where none is required - e.g. times *)
let rec sym_analogy2 source_symbs target_symbs insymbs =
  try
    if (List.length insymbs) > max_gen
    then
      raise (T_enough max_gen)
    else
      match target_symbs with
	| [] ->
	  insymbs
	| h::t ->
	  let ans = find_equiv [] insymbs h source_symbs in
(*	  (List.iter (fun (x,y) -> print_string ((pretty_func x) ^" <----> "^(pretty_func y)^"\n")) ans);*)
	  sym_analogy2 source_symbs t (ans::insymbs)
  with
      T_enough _ ->
	insymbs


let rec remcommon symbs = 
  match symbs with
    | [] ->
      ([],[])
    | (h::t) ->
      if (List.mem h common_funcs)
      then
	((fst (remcommon t)),(h::(snd (remcommon t))))
      else
	((h::(fst (remcommon t))),(snd (remcommon t)))
       

let rec istn num l = 
  if (num <= 0) 
  then 
    []
  else
    match l with
      |	[] ->
	  []
      | h::t ->
	h::(istn (num-1) t)

(* %%% reconstruct tree + test *)
let apply_symb_pair s f va = 
  let rec apply_symb_pair_term ss t v b = 
    let rec lookup a aes =
      match aes with
	| [] -> 
          (* replacing not found symbol with itself *)
	  a
	| (h,h2)::t ->
	  if (h=a) 
	  then 
	    h2
	  else
	    lookup a t
    in
    match t with
      | T_atom x ->
	[T_atom x]
      (* need to figure out what to do with atoms *)
      | T_var x ->
	if b then
	  List.append (List.map (fun x-> (T_func(Minus, [x;(T_atom 1)]))) v) v
	else 
	  [T_var x]
      | T_func(x,args) ->
	(* here going to get a list of possible return states *)
	(* return list of permuted arguments *)
	let fn = 
	  lookup x ss in
(*	print_string ((pretty_func x)^"::==::"^(pretty_func fn)^"\n"); *)
 (* calc here what to do with funcs with more or less arguments *)
	let ls = get_func_types_t x in
	let lt = get_func_types_t fn in	
	let lens = List.length ls in
	let lent = List.length lt in
	let v2 = (gen_vars v (lent-lens)) in
	(* List.iter (fun x -> (print_string ((term_pretty x)^" "))) v2; *)
	let args_list =
	      (* permute args *)
	  if (lens = lent) 
	  then	   
	    [args]
	  else 
	    (if (lens > lent)
	     (* delete arg *)
	     then 
		if b then		 
		   permute_multi args lent
		else		 
		  [istn lent args]		   
	     (* analyse types here?? *)
	     (* add arg *)
	     else 
(* adding enough args here ???? *)
	(*	(permute args))  *)

		
		(* List.iter (fun x -> print_string (term_pretty x)) v2; *)
		(* print_string "  ::   "; *)
		(* (List.iter (fun x -> print_string ((term_pretty x)^" ")) v2); *)
		(* print_string "done\n"; *)
		(* HEURISTIC CHOICE HERE - NEEDS TO BE NEW VAR *)
		if b then
		  (
		   let ans2 = permute_multi (remdup (List.append v args)) lent in
		  
		   ans2)
		else
		  [(List.append v2 args)]
	    (* List.iter (fun x -> ((List.iter (fun y -> print_string (term_pretty y)) x);print_string " ")) ans; *)
	    )
  

	(* default type correct value [args]*)
	(* (List.length lh) < (List.length h2) *)
	(* analyse types and add arg *)
	in 
if b then
  (
  (* (List.iter (fun y -> (List.iter (fun x -> print_string ((term_pretty x)^" ")) y);print_string "\n") args_list; *)

  let xx =  (List.fold_left (fun z g -> (List.append (pa_simp  (List.map (fun zz -> (apply_symb_pair_term ss zz v b)) g)) z)) [] args_list) in

  (* List.iter (fun y -> (List.iter (fun x -> print_string ((term_pretty x)^" ")) y);print_string "\n") xx;  *)
  List.map (fun xz -> T_func(fn,xz)) xx )
    (* (List.fold_left (fun z g -> (List.append (pa  (List.map (fun zz -> apply_symb_pair_term ss zz (List.append v v2) b) g)) z)) [] args_list) *)
   else
    List.map (fun xx -> T_func(fn,xx)) 
      (List.fold_left (fun z g -> (List.append (pa_simp  (List.map (fun zz -> apply_symb_pair_term ss zz (List.append v v2) b) g)) z)) [] args_list)
  (*(List.map (fun x -> (List.map (fun z -> apply_symb_pair_term ss z) x)) args_list)*)
  in
  match f with
(* classes here *)
    | F_comparison(Equals, args) ->
      match args with
	| a1::a2::[] ->
	  let funcsl = apply_symb_pair_term s a1 va false in
	  let vb = remdup (get_vars_term (List.hd funcsl)) in
	  let funcsr = apply_symb_pair_term s a2 vb true in
	  (* This is where we should do congruence closure *)
	  

	 
	  List.flatten (List.map (fun x -> (List.map (fun y -> (F_comparison(Equals,x::y::[]))) funcsr)) funcsl)
      (* F_comparison(Equals, *)
      (* 		   List.map (fun x -> apply_symb_pair_term s x) args) *)

(* let rec k s l =  *)
(*   let rec kk i m = *)
(*     match i with *)
(*       | [] ->  *)
(* 	m *)
(*       | h::t -> *)
(* 	kk t (apply_symb_pair h m) *)
(*   in *)
(*   match s with *)
(*     | [] -> *)
(*       [] *)
(*     | h::t -> *)
(*       (kk h l)::(k t l) *)

let rec print_pairs l =
  match l with
    | [] ->
       print_string "finished\n";
    | (a,b)::t ->
      print_string ((pretty_func a)^":::"^(pretty_func b)^"\n");
      print_pairs t

let perm_form x = 
  let rec all_vars x = 
    match x with
      | (T_var _)::t ->
	all_vars t
      | [] -> 
	true
      | _::t ->
	false
  in
  let rec combine x = 
    let comb_term x y =
      T_func(Plus,[x;y])
    in
    match x with
      | [] ->
	[]
      | h::t ->
	List.append
	  (List.map (fun x -> comb_term h x) t)
	  (combine t)
  in
  match x with
    | F_comparison(Equals,t1::t2::[])::_ ->
      match t1 with
	| T_func(_,l) ->
	  if (all_vars l) then 
	    (let l1 = List.map (fun x -> (match x with 
	      | F_comparison(Equals,(a::b::[])) -> b)) x
	     in
	     List.map (fun x -> F_comparison(Equals,(t1::x::[]))) (combine l1))
	  else
	    (let l1 = List.map (fun x -> (match x with 
	      | F_comparison(Equals,a::b::[]) -> a)) x
	     in
	     List.map (fun x -> F_comparison(Equals,x::t2::[])) (combine l1))
	| _ ->
	  x
  

  


let mutate source target lemma = 
  let source_symbs = collect source in
  let target_symbs = collect target in
  let lemma_symbs = collect lemma in
  (* get var types here in general*)
  let t_symbs = remdup  (List.append lemma_symbs source_symbs) in
  (*let (total_symbs,com_symbs) = remcommon t_symbs in*)
  (*  let total_symbs = List.append total_symbs2 common_funcs in  *)
  let vars_list = remdup (get_vars lemma) in
 (* doing this up front is inefficient *)
  let tss = t_symbs in 
  (* List.iter (fun x -> print_string ("tar:"^(pretty_func x)^"  ")) (remdup (List.append common_funcs target_symbs)); *)
  (* List.iter (fun x -> print_string ("src:"^(pretty_func x)^"  ")) tss; *)
  (* List.iter (fun x -> print_string ((pretty_func x)^"  ")) (List.append com_symbs target_symbs); *)
  let symb_pairs = sym_analogy2 tss (permute_multi_all (remdup (List.append common_funcs target_symbs)) (List.length tss) [] []) [] in
  (* let total_symb_pairs = apply_common com_symbs symb_pairs in *)
  (*  let symb_pairs = sym_analogy source_symbs (merge target_symbs lemma_symbs) [] in*)
  (*List.iter print_pairs symb_pairs; *)
  (* List.flatten (List.map (fun x -> apply_symb_pair x lemma vars_list) symb_pairs) *)
  let res2 =
    (
    List.flatten (List.map (fun x -> apply_symb_pair x lemma vars_list) symb_pairs) 
   ) in
  let res = remdup res2 in
  let ans = (List.fold_left (fun (x1,x2) y -> match (test y) with
    | (true,_,_) ->
      (y::x1),x2
    | (false,_,_) ->
      x1,(y::x2))
	       ([],[]) res)
  in match ans with
    | [],x ->
      let ans2 = perm_form x in
      let ans3 = (List.fold_left (fun (x1,x2) y -> match (test y) with
    | (true,_,_) ->
      (y::x1),x2
    | (false,_,_) ->
      x1,(y::x2))
		    ([],[]) ans2) in
      (match ans3 with
	| [],x -> x
	| x,_ -> x)
    | x ,_-> x
      
(*  *)
let exp_tree x = 
  let exp_t x =
    T_func(Plus,[x;x])
  in
  match x with
    |  F_comparison(Equals, [arg1;arg2]) ->
      F_comparison(Equals, [arg1;(exp_t arg2)])



let mutate_step source target lemma  = 
  let source_symbs = collect source in
  let target_symbs = collect target in
  let lemma_symbs = collect lemma in
  (* get var types here in general*)
  let t_symbs = remdup  (List.append lemma_symbs source_symbs) in
  (*let (total_symbs,com_symbs) = remcommon t_symbs in*)
  (*  let total_symbs = List.append total_symbs2 common_funcs in  *)
  let vars_list = remdup (get_vars lemma) in
 (* doing this up front is inefficient *)
  let tss = t_symbs in 
  (* List.iter (fun x -> print_string ("tar:"^(pretty_func x)^"  ")) (remdup (List.append common_funcs target_symbs)); *)
  (* List.iter (fun x -> print_string ("src:"^(pretty_func x)^"  ")) tss; *)
  (* List.iter (fun x -> print_string ((pretty_func x)^"  ")) (List.append com_symbs target_symbs); *)
  let symb_pairs = sym_analogy2 tss (permute_multi_all (remdup (List.append common_funcs target_symbs)) (List.length tss) [] []) [] in
  (* let total_symb_pairs = apply_common com_symbs symb_pairs in *)
  (*  let symb_pairs = sym_analogy source_symbs (merge target_symbs lemma_symbs) [] in*)
  (*List.iter print_pairs symb_pairs; *)
  (* List.flatten (List.map (fun x -> apply_symb_pair x lemma vars_list) symb_pairs) *)
  let lemma_exp1 = exp_tree lemma in
  let res = 
    remdup 
      (List.flatten (List.map (fun x -> apply_symb_pair x lemma_exp1 vars_list) symb_pairs))
     in
  List.fold_left (fun x y -> List.append (match (test y) with
    | (true,_,_) ->
      [y]
    | (false,_,_) ->
      [y])
    x)
    [] res
