open Eval_nats
(*open Syntax*)


type test_case = 
    (* test case type *)
  | Test_case of (term * term) list

type test_cases = 
  | Test_cases of (term * types) list

exception Test_fail of test_case
exception Test_finished



let test_pretty t =
  match t with
    | Test_case l ->
      List.fold_right
  	(fun (T_var v, T_atom i) y ->
  	  (v^": "^(string_of_int i)^", ")^y)
  	    l ""



(* test simplest case - a+b = b+a and a-b = b-a *)
(* write evaluator here *)
let print_tc tc = 
  match tc with
    | Test_case(x) ->
      List.iter (fun (T_var y,T_atom z) ->
	print_string ("Var: "^y^" "^(string_of_int z)^"\n")) x


let rec test_form tc form =
  let rec test_term tc term =
    let rec get_assoc v exsa =
      match exsa with
	| Test_case([]) ->
	  failwith "T2: assoc not found"
	| Test_case((a,ex)::t) ->
	  if (a = v)
	  then
	    ex
	  else 
	    (get_assoc v (Test_case(t)))
    in
    match term with
	(* leaf case *)
      | T_var x -> 
	get_assoc (T_var x) tc
	  (* functions *)
      | T_atom x ->
	(T_atom x)
      | T_func(func,args) ->
	T_func(func,List.map (fun x-> test_term tc x) args)
  in
  match form with 
    | F_comparison(Equals,(t1::t2::[])) ->
      eval(F_comparison(Equals,(test_term tc t1)::(test_term tc t2)::[]))

let print_tcs tcs =
  List.iter
    (fun (T_var x,l) ->
      (print_string ("\nVar "^x^":");
       (List.iter (fun x ->
	 (print_string ((string_of_int x)^" ")))
	   l))) tcs


(* get next testcase *)
(* should never be all empty *)
let rec calc_next tcs tcs_orig = 
  let rec find_tc (T_var v) t_o = 
    match t_o with
      | [] ->
	failwith "T2: Var not found"
      | (T_var a,l)::t_ol ->
	if (v=a)
	then
	 (* (print_string ("Var v:"^v^", Var a:"^a^" ");
	   List.iter (fun x -> print_string ((string_of_int x)^" ")) l;*)
	  ((T_var a),l)
	else
	  find_tc (T_var v) t_ol
  in
  match tcs with 
    | (_,[_])::[] ->
      raise Test_finished
    | (var,h1::h2::t)::tl ->
      (var,h2::t)::tl
    | (var,[_])::tl ->
      (find_tc var tcs_orig)::(calc_next tl tcs_orig)
    | (_,[])::_ ->
      failwith "next calc failed (T4)"

let rec test ans tcst tcsf all_ex expr tcs_orig tcs = 
  try
    let ntcs = calc_next tcs tcs_orig in
    let tc = Test_case (List.map (fun (x,y) ->
      (x,(T_atom (List.hd y)))) tcs)
    in 
    (* (match tc with *)
    (*   | Test_case(tcc) -> *)
	(*List.iter (fun (T_var v,T_atom x) ->*)
	(*  print_string ("Var "^v^": "^(string_of_int x)^"\n")) tcc);*)
  (* exception handle here *)
    (try
      match ntcs with
	| [] ->
	  ans,tcst,tcsf
	| _ ->
	  if (test_form tc expr)
	  then
	     test ans (tc::tcst) tcsf all_ex expr tcs_orig ntcs
	  else
	    raise (Test_fail tc)
    with Test_fail tc ->
      if all_ex
      then
	test false tcst (tc::tcsf) all_ex expr tcs_orig ntcs
      else
	false,[],[tc])
  with
      Test_finished ->
	ans,tcst,tcsf
	  
(* type inference here *)
(* if type is given then recover that *)
(* if not then try and propagate types *)

(**** TODO - not treating functions properly ***)


let rec get_types expr exp_type =
  let rec get_types_term expr_t exp_type_t =
  match expr_t with
    (* end_var then add the type so far inferred *)
    |T_atom x ->
      []
    | T_var x ->
      (T_var x,exp_type_t)::[]
    | T_func(func,args) ->
      let func_types = get_func_types_t func in
(* TODO _ RESULT TYPE + HIGHER ORDER??? *)
       (List.fold_left (fun x (y,z) -> (List.append (get_types_term y z) x)) 
       	[] (List.map2 (fun x y -> (x,y)) args func_types)) 
  in
  match expr with
    | F_comparison(Equals,(arg1::arg2::[])) ->
      List.append
	(get_types_term arg1 Tcs)
	(get_types_term arg2 Tcs)

(* let rec print_tcs t =  *)
(*   match t with *)
(*     | [] -> *)
(*       Printf.printf "\n"; *)
(*     | (T_var v,ints)::t -> *)
(*       List.iter (fun x -> Printf.printf ("Var ")) ints; *)
(*       Printf.printf "\n"; *)
(*       print_tcs t *)

(* argument from input parse goes here *)
let test expr =
  (* parse string*)
  (* let expr = F_comparison(Equals,(T_func(Plus,(T_var "y")::(T_var "x")::[])::T_func(Plus,(T_var "x")::(T_var "y")::[])::[])) (\* parse string for now put expr here *\)  *)
  (* in *)
  let typs = get_types expr T_bool in
(* remove duplicates !!!!!!*)
  (* typs is list [(var,type)] *)
  let tcs = get_test_cases typs [] in
(* tcs is list (var,[1,2,3,4]),(var,[1,2,3],[4,5,6]) etc *)
(* second argument is whether or not it gives back all examples or not *)
  test true [] [] false expr tcs tcs

let thms_lems = 
  ((F_comparison(Equals,(T_func(Helper_fact,(T_var "n")::(T_atom 1)::[])::(T_func(Theta_fact,T_var("n")::[]))::[]))),
   (F_comparison(Equals,
		(T_func(Helper_fact,(T_var "n")::(T_var "a")::[])::
		   (T_func(Times,(T_var "a")::(T_func(Theta_fact,(T_var "n")::[]))::[]))::[]))))::
    ((F_comparison(Equals,(T_func(Helper_power,(T_var "n")::(T_atom 1)::[])::(T_func(Theta_power,T_var("n")::[]))::[]))),
     (F_comparison(Equals,
		   (T_func(Helper_power,(T_var "n")::(T_var "a")::[])::
		      (T_func(Times,(T_var "a")::(T_func(Theta_power,(T_var "n")::[]))::[]))::[]))))::
    (F_comparison(Equals,(T_func(Helper_expt,(T_var "m")::(T_var "n")::(T_atom 1)::[])::(T_func(Theta_expt,T_var("m")::T_var("n")::[]))::[])),
     (F_comparison(Equals,
		   (T_func(Helper_expt,(T_var "m")::(T_var "n")::(T_var "a")::[])::
		      (T_func(Times,(T_var "a")::(T_func(Theta_expt,(T_var "m")::(T_var "n")::[]))::[]))::[]))))::
    (F_comparison(Equals,(T_func(Helper_mul,(T_var "m")::(T_var "n")::(T_atom 1)::[])::(T_func(Theta_mul,T_var("m")::T_var("n")::[]))::[])),
  (F_comparison(Equals,
		   (T_func(Helper_mul,(T_var "m")::(T_var "n")::(T_var "a")::[])::
		      (T_func(Plus,(T_var "a")::(T_func(Theta_mul,(T_var "m")::(T_var "n")::[]))::[]))::[]))))::
  ((F_comparison(Equals,(T_func(Helper_sum_square,(T_var "n")::(T_atom 0)::[])::(T_func(Theta_sum_square,T_var("n")::[]))::[]))),
   (F_comparison(Equals,
		(T_func(Helper_sum_square,(T_var "n")::(T_var "a")::[])::
		   (T_func(Plus,(T_var "a")::(T_func(Theta_sum_square,(T_var "n")::[]))::[]))::[]))))::
  ((F_comparison(Equals,(T_func(Helper_sum,(T_var "n")::(T_atom 0)::[])::(T_func(Theta_sum,T_var("n")::[]))::[]))),
   (F_comparison(Equals,
		(T_func(Helper_sum,(T_var "n")::(T_var "a")::[])::
		   (T_func(Plus,(T_var "a")::(T_func(Theta_sum,(T_var "n")::[]))::[]))::[]))))::
    ((F_comparison(Equals,(T_func(Helper_fib,(T_var "n")::(T_atom 0)::(T_atom 1)::[])::(T_func(Theta_fib,T_var("n")::[]))::[])))
      ,(F_comparison(Equals,
(T_func(Helper_fib,(T_var "n")::(T_var "m")::(T_var "a")::[]))::
(T_func(Plus,(T_func(Times,(T_func(Theta_fib,(T_func(Minus,(T_var "n")::(T_atom 1)::[]))::[]))::(T_var "m")::[]))::(T_func(Times,(T_func(Theta_fib,(T_var "n")::[]))::(T_var "a")::[]))::[]))::[])))::    
    []
   
let target_thm2 = 
  F_comparison(Equals,(T_func(Helper_power,(T_var "n")::(T_atom 1)::[])::(T_func(Theta_power,T_var("n")::[]))::[]))

let target_thm = 
  F_comparison(Equals,(T_func(Helper_expt,(T_var "m")::(T_var "n")::(T_atom 1)::[])::(T_func(Theta_expt,T_var("m")::T_var("n")::[]))::[]))

let input_term =
  (F_comparison(Equals,
		(T_func(Helper_fact,(T_var "n")::(T_var "a")::[])::
		   (T_func(Times,(T_var "a")::(T_func(Theta_fact,(T_var "n")::[]))::[]))::[])))



let input_terms =
  (F_comparison(Equals,
		(T_func(Helper_fact,(T_var "n")::(T_var "a")::[])::
		   (T_func(Times,(T_var "a")::(T_func(Theta_fact,(T_var "n")::[]))::[]))::[])))::
    (F_comparison(Equals,
		(T_func(Helper_power,(T_var "n")::(T_var "a")::[])::
		   (T_func(Times,(T_var "a")::(T_func(Theta_power,(T_var "n")::[]))::[]))::[])))::
    (F_comparison(Equals,
		(T_func(Helper_power,(T_var "a")::(T_var "n")::[])::
		   (T_func(Times,(T_var "a")::(T_func(Theta_power,(T_var "n")::[]))::[]))::[])))::
    (F_comparison(Equals,
		(T_func(Helper_fib,(T_var "n")::(T_var "m")::(T_var "a")::[])::
		   (T_func(Plus,
			   (T_func(Times,(T_func(Theta_fib,(T_func(Minus,(T_var "n")::(T_atom 1)::[]))::[]))::(T_var "m")::[]))::
			     (T_func(Times,(T_func(Theta_fib,(T_var "n")::[]))::(T_var "a")::[]))
		    ::[])
		    )::[])))::
    []


    
