(*---------------------------------------------------------------------------
   Copyright (c) 2007-2012 Daniel C. Bünzli. All rights reserved.
   Distributed under a BSD3 license, see license at the end of the file.
  ---------------------------------------------------------------------------*)
open Eval_nats
open Mutation
open Tester
open Types

exception Invalid_string

let str = Printf.sprintf
let exec = Filename.basename Sys.executable_name
let pr_err s = Printf.eprintf "%s:%s\n" exec s
let apply f x ~finally y = 
  let result = try f x with exn -> finally y; raise exn in
  finally y;
  result

let fail ((l, c), e) = failwith (str "%d:%d: %s" l c (Xmlm.error_message e))

type tree = E of Xmlm.tag * tree list | D of string

let in_tree i = 
  let el tag childs = E (tag, childs)  in
  let data d = D d in
  Xmlm.input_doc_tree ~el ~data i

(* data represents tabs and spaces too if I can be bothered *)
(* let addinfo_orig = *)
(*   List.flatten (List.map (fun (r,n,l) ->   *)
(*     ((D "    ")::E((("","relevance"),[]),(D r)::[]):: *)
(*       (D "\n        ")::E((("","uses"),[]), *)
(* 	(D "\n            ")::E((("","lemma"),[]),  *)
(*      	  (D "\n                ")::E((("","name"),[]),(D n)::[])::(D "\n                ")::E((("","statement"),[]),(D "\n                    ")::(D l)::(D "\n                ")::[])::(D "\n            ")::[])::(D "\n        ") *)
(* 	::[])::(D "\n    ")::[])) *)
(*   (("1","lem1","test_st1")::("2","lem2","test_st2")::[])) *)

let addinfo l = 
  let p = List.map
    (fun (a,b) ->
      E((("","lemma"),[]),
	 (E((("","name"),[]),(D a)::[])::
	    E((("","statement"),[]),(D b)::[])::[])))
(*	[("lem1","test_st1");("lem2","test_st2")]*)
    l
  in
  E((("","uses"),[]),p)::[]


(* to do then - 
   extract relevance n source and lemma - parse strings 
   
*)

let rec preproc ds (tt,a) = 
  let rec preproc_inner dsl =
    List.map
      (fun x ->
	match x with
	  | E(((_,"source-conjecture"),_),
	      E(((_,"name"),_),(D _)::[])::
		 E(((_,"statement"),_),(D st)::[])::
		 E(((_,"relevance"),_),(D _)::[])::
		 E(((_,"uses"),_),
		   E(((_,"lemma"),_),
		     E(((_,"name"),_),(D _)::[])::
		       E(((_,"statement"),_),(D sl)::[])::[])::[])::[]) ->
	      (st,sl)
      )
      dsl
  in
  match ds with
    | E(((_,"target-conjecture"),_),E(((_,"name"),_),(D _)::[])::
	E(((_,"statement"),_),(D d)::[])::[])::t ->
      preproc t (d,a)
    | E(((_,"source-conjectures"),_),
	l)::_ ->
      (tt,preproc_inner l)
    | _ ->
      (tt,a)
    
let preproc2 d =
  match d with
    | E(_,t) ->
(*      print_string "1";*)
      preproc t ("",[])
    | D _ ->
  (*    print_string "2";*)
      ("",[])


(* let find str sub = *)
(*         let sublen = length sub in *)
(*         if sublen = 0 then *)
(*                 0 *)
(*         else *)
(*                 let found = ref 0 in *)
(*                 let len = String.length str in *)
(*                 try *)
(*                         for i = 0 to len - sublen do *)
(*                                 let j = ref 0 in *)
(*                                 while String.unsafe_get str (i + !j) = String.unsafe_get sub !j do *)
(*                                         incr j; *)
(*                                         if !j = sublen then begin found := i; raise Exit; end; *)
(*                                 done; *)
(*                         done; *)
(*                         raise Invalid_string *)
(*                 with *)
(*                         Exit -> !found *)

let rec d x = 
  match x with
    | Atom x ->
      ("Atom \""^x^"\"")
    | Cons(x,y) ->
      ("Cons("^(d x)^","^(d y)^")")

let rec testtranslate_acl2_2 x = 
  (* print_string (d x); *)
  (* print_string "\n\n"; *)
  match x with
    | Cons((Atom "fn_fact"),(Cons((Atom a),(Atom "nil")))) ->
      T_func(Helper_fact,(T_var a)::(T_atom 1)::[])
    | Cons((Atom "theta_fact"),(Cons((Atom a),(Atom "nil")))) ->
      T_func(Theta_fact,(T_var a)::[])
    | Cons((Atom "fn_fib"),(Cons((Atom a),(Atom "nil")))) ->
      T_func(Helper_fib,(T_var a)::(T_atom 1)::(T_atom 0)::[])
    | Cons((Atom "theta_fib"),(Cons((Atom a),(Atom "nil")))) ->
      T_func(Theta_fib,(T_var a)::[]) 
    | Cons(Atom "*",Cons(Atom a,Cons(Cons(Atom "theta_fact",Cons(Atom n,Atom "nil")),Atom "nil"))) ->
      T_func(Times,(T_var a)::(T_func(Theta_fact,(T_var n)::[]))::[])
    | Cons(Atom "helper_fact",Cons(Atom n,Cons(Atom a,Atom "nil"))) ->
      T_func(Helper_fact,(T_var n)::(T_var a)::[])
    | _ -> failwith "not known"

let rec translate_acl2_2 x =
  let rec bf l =
    match l with
      | Atom "nil" ->
	[]
      | Cons(x, y) ->
	(translate_acl2_2 x)::(bf y)
  in
  match x with
    | Atom "1" ->
      (T_atom 1)
    | Atom "0" ->
      (T_atom 0)
    | Cons((Atom "fn_power"),l) ->
      T_func(Helper_power,(List.append (bf l) ((T_atom 1)::[])))
    | Cons((Atom "theta_power"),l) ->
      T_func(Theta_power,bf l)
    | Cons((Atom "helper_power"),l) ->
      T_func(Helper_power,bf l)
    | Cons((Atom "fn_mul"),l) ->
      T_func(Helper_mul,(List.append (bf l) ((T_atom 1)::[])))
    | Cons((Atom "theta_mul"),l) ->
      T_func(Theta_mul,bf l)
    | Cons((Atom "helper_mul"),l) ->
      T_func(Helper_mul,bf l)
    | Cons((Atom "fn_expt"),l) ->
      T_func(Helper_expt,(List.append (bf l) ((T_atom 1)::[])))
    | Cons((Atom "theta_expt"),l) ->
      T_func(Theta_expt,bf l)
    | Cons((Atom "helper_expt"),l) ->
      T_func(Helper_expt,bf l)
    | Cons((Atom "fn_sum"),l) ->
      T_func(Helper_sum,(List.append (bf l) ((T_atom 0)::[])))
    | Cons((Atom "theta_sum"),l) ->
      T_func(Theta_sum,bf l)
    | Cons((Atom "helper_sum"),l) ->
      T_func(Helper_sum,bf l)
    | Cons((Atom "fn_sum_square"),l) ->
      T_func(Helper_sum_square,(List.append (bf l) ((T_atom 1)::[])))
    | Cons((Atom "theta_sum_square"),l) ->
      T_func(Theta_sum_square,bf l)
    | Cons((Atom "helper_sum_square"),l) ->
      T_func(Helper_sum_square,bf l)
    | Cons((Atom "fn_fact"),l) ->
      T_func(Helper_fact,(List.append (bf l) ((T_atom 1)::[])))
    | Cons((Atom "theta_fact"),l) ->
      T_func(Theta_fact,bf l)
    | Cons((Atom "fn_fib"),l) ->
      T_func(Helper_fib,(List.append (bf l) ((T_atom 1)::(T_atom 0)::[])))
    | Cons((Atom "theta_fib"),l) ->
      T_func(Theta_fib,(bf l))
    | Cons((Atom "helper_fact"),l) ->
      T_func(Helper_fact,(bf l))
    | Cons((Atom "+"),l) ->
      T_func(Plus, (bf l))
    | Cons((Atom "-"),l) ->
      T_func(Minus, (bf l))
    | Cons((Atom "*"),l) ->
      T_func(Times,(bf l))
    | Cons(x,(Atom "nil")) ->
      (translate_acl2_2 x)
    | Atom x ->
      (T_var x)
    

	


let rec translate_acl2 x = 
  match x with 
    | Cons((Atom "defthm"),Cons((Atom name),
			 Cons(Cons((Atom "implies"),Cons(_,
						  Cons(Cons((Atom "equal"),
							    (Cons(l,(Cons(r,(Atom "nil")))))
							    (*Cons(Cons("fn_fib",Cons("n","nil")),Cons(Cons("theta_fib",Cons("n","nil")),"nil"))*)
						  )
							 ,(Atom "nil")))),(Atom "nil"))))
	->
      let ans = (name,F_comparison(Equals,(translate_acl2_2 l)::(translate_acl2_2 r)::[])) in
      (*print_string (form_pretty (snd ans));*)
      ans
    | _ -> failwith "not parsable"


(*let rec translate_acl2 x =
  match x with
    | Atom x ->
      ("\""^x^"\"")
    | Cons(x,y) ->
      ("Cons("^(translate_acl2 x)^","^(translate_acl2 y)^")")*)


let translate x =
  (*F_comparison(Equals,(T_var "x")::(T_var "y")::[])*)
  let lex = Lexing.from_string x in
  let ast = Parser.main Lexer.token lex in
  translate_acl2 ast
(*  ("a",F_comparison(Equals,(T_var "x")::(T_var "y")::[]))*)
  
let rec ttback ss = 
  let rec ttback_t s2 = 
    match s2 with
      | T_func(f,l) ->
	"(" ^(pretty_func f)^" "^(List.fold_left 
				    (fun a b -> (a^(ttback_t b)^" ")) "" l)^")"
      | T_atom n ->
	string_of_int n
      | T_var x ->
	x
  in
  match ss with
    | F_comparison(Equals,t1::t2::[]) ->
      "(equal "^(ttback_t t1)^" "^(ttback_t t2)^")"


let translate_back name x =
  let vars = remdup (get_vars x) in
  let s = 
    (
      match vars with
	| [T_var v] -> "(natp "^v^")"
	| _ ->
	  "(and "^
	   (List.fold_left (fun a b -> (let x = 
					    (match b with 
					      | (T_var x) -> x) in
					  (a^("(natp "^x^") ")))) "" vars) ^")"
    )
  in
  let xx = "(defthm "^name^" (implies "^s^" "^(ttback x)^"))" in
  (* need to get vars and write out *)
      (name, xx)
(* need to add in the preamble *)
(* also need to work out how to add in fn_fib etc... *)
    
    
let mut name tt st sl =  
  let ts = mutate st tt sl  in
  let tts = (match ts with
    | [] -> [sl]
  (* let tts = (match ts with  *)
  (*   | [] -> *)
  (*     (let ttts = mutate_step st tt sl  in *)
  (*      match ttts with *)
  (* 	 | [] -> *)
  (* 	   [sl] *)
  (* 	 | _ -> *)
  (* 	   ttts *)
   (*   )    *)
    | _ ->
      ts) 
  in
  (List.map (fun x ->
    (translate_back name x))
     tts)
      

let get_info tt l =
  let (name,target_theorem) = tt in
  (List.flatten (List.map (fun (x,y) ->
    let (name2,xx) = (translate x) in
     mut (name^"_FROM_"^name2) target_theorem xx (snd (translate y)))
    l))
    


let rec proc ds p =
  match ds with
    | D d ->
      D d
    | E(((a,b),c),d) ->
      if b = "target-conjecture" 
      then
	E(((a,b),c),(List.append (List.map (fun x -> proc x p) d) (addinfo p)))
      else
	E(((a,b),c),List.map (fun x -> proc x p) d)

let out_tree o t = 
  let pp l = 
    List.map (fun x ->
      match x with 
	| ((n1,n2),a)
	    -> ((("qq"^n1),("ll"^n2)),("pp"^a))
    )
      l
  in
  let rec man ds =
    match ds with
      | D d ->
	D d
      | E (a,b) ->
	E (a,b) 
  in
  (* E(((n1,n2),l),c) -> *)
  (* print_string ("n1: "^n1^" n2: "^n2^"\n l: "); *)
  (* List.iter (fun ((a,b),c) -> (print_string ("la: "^a^" lb:"^b^" lc:"^c^"\n"))) l; *)
  (* E((((n1),(n2)),(l)),(List.map man c)) *)
  let mon a = 
    match a with
      | None -> None
      | Some d -> Some d 
  in
  let prepr a = 
    match a with 
      | (t,d) ->
	((mon t), (man d))
  in
  let frag = function
      | E (tag, childs) -> `El (tag, childs) 
      | D d -> `Data d 
  in
  let (ans_t,ans_l) = (prepr t) in
  let (tt,sls) = preproc2 ans_l in
  let res = get_info (translate tt) sls in
  (*List.iter (fun (x,y) -> print_string (x^y)) sls;*)
  (*let ans1 = List.map (fun (x,y) -> (("SOURCE THEOREM: "^x),("SOURCE LEMMA: "^y))) sls in*)
  (*print_string ("TARGET THEOREM: "^tt);*)
  (* enumerate lems here ?? *)
  let ans2 = proc ans_l res in
  Xmlm.output_doc_tree frag o (ans_t,ans2)
    

let xml_parse tree enc strip entity ns ic () =                (* parse only *)
  let i = Xmlm.make_input ~enc ~strip ~entity ~ns (`Channel ic) in
  let doc i = 
    if tree then ignore (in_tree i) else
    begin 
      let rec pull i l = match Xmlm.input i with
      | `El_start _ -> pull i (l + 1)
      | `El_end -> if l = 1 then () else pull i (l - 1)
      | `Data _ -> pull i l 
      | `Dtd _ -> assert false
      in
      ignore (Xmlm.input i); (* `Dtd *)
      pull i 0;
    end
  in
  try while not (Xmlm.eoi i) do doc i done 
  with Xmlm.Error (p, e) -> fail (p, e)

let xml_signals _ enc strip entity ns ic _ =           (* output signals *)
  let pr s = Printf.fprintf stdout s in
  let i = Xmlm.make_input ~enc ~strip ~entity ~ns (`Channel ic) in 
  let out_signal s = match s with
  | `Dtd None -> pr "`DTD None\n"
  | `Dtd Some d -> pr "`DTD Some(%S)\n" d
  | `El_start (n, atts) -> 
      let pr_name (p, l) =  if p <> "" then pr "%s:%s" p l else pr "%s" l in
      let pr_att (n, v) = pr "("; pr_name n; pr ", %S); " v in
      pr "`El_start (";
      pr_name n; pr ", ["; List.iter pr_att atts; pr "])\n"
  | `El_end -> pr "`El_end\n"
  | `Data d -> pr "`Data (%S)\n" d
  in
  try while not (Xmlm.eoi i) do out_signal (Xmlm.input i); done
  with Xmlm.Error (p, e) -> fail (p, e)
  
let xml_outline tree enc strip entity ns ic oc =            (* ascii outline *)
  let pr s = Printf.fprintf oc s in
  let pr_dtd dtd = match dtd with Some s -> pr "+-DTD %S\n" s | _ -> () in
  let pr_depth d = for k = 1 to d do pr "| " done in
  let pr_data d data = pr_depth d; pr "%S\n" data in
  let pr_name c (p, l) =  if p <> "" then pr "%s:%s" p l else pr "%s" l in
  let pr_att d (n, v) = pr_depth (d + 1); pr "* %a = %S\n" pr_name n v in
  let pr_tag d (n, atts) = 
    pr_depth d; pr "+-%a\n" pr_name n; List.iter (pr_att d) atts 
  in
  let i = Xmlm.make_input ~enc ~strip ~entity ~ns (`Channel ic) in
  let doc i = 
    if tree then 
      begin 
	let rec pr_tree d = function
	  | (n :: next) :: path -> 
	      begin match n with
	      | D data -> pr_data d data; pr_tree d (next :: path)
	      | E (tag, childs) -> 
		  pr_tag d tag; pr_tree (d+1) (childs :: next :: path)
	      end
	  | [] :: path -> if d = 0 then () else pr_tree (d - 1) path
	  | _ -> assert false
	in
	let dtd, t = in_tree i in
	pr_dtd dtd;
	pr_tree 0 ([t] :: [])
      end
    else
      begin 
	let rec pull i l = match Xmlm.input i with
	| `El_start tag -> pr_tag l tag; pull i (l + 1)
	| `El_end -> if l = 1 then () else pull i (l - 1)
	| `Data d -> pr_data l d; pull i l
	| `Dtd _ -> assert false 
	  in
	pr_dtd (match Xmlm.input i with `Dtd d -> d | _ -> assert false);
	pull i 0;
      end;
    flush oc
  in
  try while not (Xmlm.eoi i) do doc i done
  with Xmlm.Error (p, e) -> fail (p, e)

let xml_xml2  indent tree enc strip entity ns ic oc =   
  let nl = (indent = None) in
  let i = Xmlm.make_input ~enc ~strip ~ns ~entity (`Channel ic) in
  let o = Xmlm.make_output ~nl ~indent ~ns_prefix:ns (`Channel oc) in
  let dstruct = in_tree i in
  out_tree o dstruct
  

let xml_xml indent tree enc strip entity ns ic oc =              (* xml trip *)
  let nl = (indent = None) in
  let i = Xmlm.make_input ~enc ~strip ~ns ~entity (`Channel ic) in
  let o = Xmlm.make_output ~nl ~indent ~ns_prefix:ns (`Channel oc) in
  let doc i o = 
    if tree then (out_tree o (in_tree i)) else
    begin 
      let rec pull i o depth = 
	let s = Xmlm.input i in
	Xmlm.output o s;
	match s with 
	| `El_start _ -> pull i o (depth + 1)
	| `El_end -> if depth = 1 then () else pull i o (depth - 1)
	| `Data _ -> pull i o depth 
	| `Dtd _ -> assert false
      in
      Xmlm.output o (Xmlm.input i); (* `Dtd *)
      pull i o 0
    end
  in
  try while not (Xmlm.eoi i) do doc i o done 
  with Xmlm.Error (p, e) -> fail (p, e)

let with_inf f inf v = 
  try
    let ic = if inf <> "" then open_in_bin inf else stdin in 
    let close ic = if inf <> "" then close_in ic else () in 
    apply (f ic) v ~finally:close ic
  with
  | Sys_error e -> pr_err (str " %s" e)
  | Failure e -> pr_err (str "%s:%s" inf e)
	
let with_outf f ic outf = 
  try 
    let oc = if outf <> "" then open_out_bin outf else stdout in 
    let close oc = if outf <> "" then close_out oc else () in 
    apply (f ic) oc ~finally:close oc
  with
  | Sys_error e -> pr_err (str " %s" e)

let entity_fun eref xhtml = 
  if not xhtml then (if eref then fun x -> Some x else fun x -> None) else
  let h = Hashtbl.create 270 in
  List.iter (fun (e, ustr) -> Hashtbl.add h e ustr) Xhtml.entities;
  if eref then (fun x -> try Some (Hashtbl.find h x) with Not_found -> Some x)
  else (fun x -> try Some (Hashtbl.find h x) with Not_found -> None)
        
let process signals tree enc strip eref ns xhtml parse_only outline indent 
            suffix files = 
  let entity = entity_fun eref xhtml in
  let ns = if ns then fun x -> Some x else fun x -> None in
  let f = 
    if parse_only then 
      fun inf -> with_inf (xml_parse tree enc strip entity ns) inf ()
    else 
      let outf inf =  
	if inf = "" || suffix = "" then "" (* stdout *) else 
	str "%s.%s" inf suffix 
      in
      let f = 
	if outline then xml_outline else 
	if signals then xml_signals else
	(xml_xml2 indent) 
      in
      fun inf -> 
	with_inf (with_outf (f tree enc strip entity ns)) inf (outf inf)
  in
  List.iter f files

let encoding_of_str enc = match (String.lowercase enc) with
| "" -> None
| "utf-8" | "utf8" | "utf_8" -> Some `UTF_8
| "utf-16" | "utf16" | "utf_16" -> Some `UTF_16
| "utf-16be" | "utf16be" | "utf16_be" -> Some `UTF_16BE
| "utf-16le" | "utf16le" | "utf16_le" -> Some `UTF_16LE
| "iso-8859-1" | "iso88591" 
| "iso_8859_1" | "latin1" | "latin-1" -> Some `ISO_8859_1 
| "ascii" | "us-ascii" -> Some `US_ASCII
| e -> pr_err (str "unknown encoding '%s', trying to guess." e); None

let main () = 
  let usage = 
    str "Usage: %s <options> <files>\n\
         Reads xml files and outputs them on stdout.\n\
         Options:" exec 
  in
  let enc = ref "" in
  let strip = ref true in
  let ns = ref false in
  let eref = ref false in
  let xhtml = ref false in
  let parse_only = ref false in
  let tree = ref true in
  let signals = ref false in
  let outline = ref false in
  let indent = ref true in
  let suffix = ref "" in
  let files = ref [] in
  let add_file s = files := s :: !files in
  let options = [
    "-enc", Arg.Set_string enc, 
    "<enc>, use specified encoding, utf-8, utf-16, utf-16be, utf-16le,\n\
    \   iso-8859-1, ascii (otherwise guesses).";
    "-strip", Arg.Set strip, 
    "strip and collapse white space in character data.";
    "-ns", Arg.Set ns,
    "replace unbound namespaces prefixes by themselves (on input and output).";
    "-eref", Arg.Set eref,
    "replace unknown entity references by their name.";
    "-xhtml", Arg.Set xhtml,
    "resolve XHTML character entities.";
    "-p", Arg.Set parse_only, 
    "parse only, no output.";
    "-t", Arg.Set tree, 
    "build document tree in memory.";
    "-signals", Arg.Set signals,
    "outputs the stream of signals instead of xml (excludes -t).";
    "-ot", Arg.Set outline, 
    "output document ascii outline instead of xml.";
    "-indent", Arg.Set indent,
    "indent xml output.";
    "-trip", Arg.Set_string suffix,
    "<suffix>, result for file <file> is output to a file <file.suffix>."; ]
  in
  Arg.parse options add_file usage;
  let files = match (List.rev !files) with [] -> ["" (* stdin *) ] | l -> l in
  let enc = encoding_of_str !enc in
  let indent = if !indent then Some 2 else None in
  process !signals !tree enc !strip !eref !ns !xhtml !parse_only 
    !outline indent !suffix files

let () = main ()

(*---------------------------------------------------------------------------
  Copyright (c) 2007-2012 Daniel C. Bünzli
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are
  met:
        
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the
     distribution.

  3. Neither the name of Daniel C. Bünzli nor the names of
     contributors may be used to endorse or promote products derived
     from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  ---------------------------------------------------------------------------*)

